;;; roll script, Copyright (C) 2011, 2012 Melikamp.

;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation, either version 3 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but without any warranty; compiling, interpreting, executing,
;;; reading, or merely musing about the text of the program may result
;;; in lapses of consciousness and/or very being, up to and including
;;; the end of all existence and the Universe as we know it.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program (most likely, a file named COPYING).  If
;;; not, see <http://www.gnu.org/licenses/>.

;;; configuration

(defparameter *random-device* "/dev/urandom"
  "default random device, file, or pipe")

(defparameter *roll-separator* ""
  "this string (with formats!) is printed in between successive rolls")

(defparameter *perm-separator* ""
  "this string (with formats!) is printed in between permuted items")

(defparameter *size* nil
  "size of permutation, nil stands for maximum size")

(defparameter *int-format* "~A"
  "Integer output format.")

;; platform parameters

(defconstant BITS_IN_BYTE 8)

;; load custom configuration

(let ((config-file (probe-file (concatenate 'string (ext:getenv "HOME")
					    "/.config/roll/config.lisp"))))
  (if config-file (load config-file)))

;;; global variables

(defparameter *roll-version* "20")
(defparameter *random-stream* nil "stream associated with random device")
(defparameter *multiplier* 1 "number of rolls")
(defparameter *roll-command* nil "the main instruction")
(defparameter *pass-alphabet*
  "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
  "default alphabet for rolling a password")
(defparameter *deck52*
  '("As" "Ah" "Ad" "Ac"
    "2s" "2h" "2d" "2c"
    "3s" "3h" "3d" "3c"
    "4s" "4h" "4d" "4c"
    "5s" "5h" "5d" "5c"
    "6s" "6h" "6d" "6c"
    "7s" "7h" "7d" "7c"
    "8s" "8h" "8d" "8c"
    "9s" "9h" "9d" "9c"
    "Ts" "Th" "Td" "Tc"
    "Js" "Jh" "Jd" "Jc"
    "Qs" "Qh" "Qd" "Qc"
    "Ks" "Kh" "Kd" "Kc"
    "52 playing card deck"))

;; live documentation
(defparameter *command-help*
  '(("b" "a Bernoulli trial")
    ("bin" "a binomial experiment")
    ("int" "an integer from the given interval")
    ("item" "an item from the given list")
    ("pass" "a password on digits and latin letters [and more]")
    ("perm" "a permutation of items from the given list")
    ("str" "a literal string on the given alphabet")
    ("version" "print version information"))
  "list with names and short descriptions of commands")

(defparameter *option-help*
  '((" " "set roll separator to a single space")
    ("if" "integer output format (~A)")
    ("rs" "separator between multiple rolls (empty string)")
    ("ps" "separator between permuted items (empty string)")
    ("n" "size of permutation (maximum)")
    ("src" "file name to use as the source of random bytes (/dev/urandom)"))
  "list with names and short descriptions of options")

;;; functions

;; Parse a rational number

(defun parse-rational (s)
  "I parse a rational number from the string s, return nil when
  fail."
  (assert (stringp s) nil "roll: parse-rational: argument is not a string.")
  (let ((r (read-from-string s)))
    (if (rationalp r) r nil)))

;; generate a fair integer

(defun rand-int (from to)
  "I return an integer uniformly distributed in the given interval."
  (let* ((diff (- to from)) (r (1+ diff)) (bit-count) (stick))
    (if (< diff 1) from
      (progn
	(loop while (> r diff) do
	      (setf r 0)
	      (setf stick 1)
	      (setf bit-count 0)
	      (loop while (<= stick diff) do
		    (incf r (ash (read-byte *random-stream*) bit-count))
		    (setf stick (ash stick BITS_IN_BYTE))
		    (incf bit-count BITS_IN_BYTE))
	      (unless (= diff (- stick 1))
		(progn
		  (loop while (> stick diff) do (setf stick (ash stick -1)))
		  (setf r (logand r (- (ash stick 1) 1))))))
	(+ from r)))))

;; Run a Bernoulli trial

(defun bernoulli-trial (p)
  "I run a Bernoulli trial with success probability p and return 1 on
success, 0 on failure. p is assumed to be a rational number."
  (if (<= (rand-int 1 (denominator p)) (numerator p)) 1 0))

;; Run a binomial trial

(defun binomial-trial (p k)
  "I run a binomial trial with success probability p and k Bernoulli
trials. p is assumed to be a rational number, k an integer."
  (loop for n from 1 to k sum (bernoulli-trial p)))

;; shift CL arguments

(defun shift-args (by)
  "shift command line arguments"
  (setf *args* (nthcdr by *args*)))

;; expand - by reading a list from STDIN

(defun expand- ()
  "if *args* start with - then they are replaced by a list of lines from STDIN"
  (if (equal "-" (car *args*))
      (progn
	(setf *args* nil)
	(setf *args* (loop for line = (read-line nil nil)
			   while line
			   collect line)))))

;; print help

(defun print-help ()
  "print online help"
  (format *error-output* "~
roll: usage: roll [<#>] [options] [<command>] ...~%~
for help with command or option <foo>, roll <foo>~%~
<#> is the number of rolls~%~
if the list of items begins with - then items are read from stdin~%")
  (format *error-output* "  commands:~%")
  (loop for x in *command-help*
	do (format *error-output* "~16A ~A~%" (car x) (cadr x)))
  (format *error-output* "  options:~%")
  (loop for x in *option-help*
	do (format *error-output* "~16A ~A~%" (car x) (cadr x)))
  (exit 0))

;; commands

(defun roll-b (p)
  "I roll a Bernoulli trial. p is assumed to be a rational number."
  (format t *int-format* (bernoulli-trial p)))

(defun roll-bin (p k)
  "I roll a binomial trial. p is assumed to be a rational number, k an
integer."
  (format t *int-format* (binomial-trial p k)))

(defun roll-int (from to)
  (format t *int-format* (rand-int from to)))

(defun roll-item (&rest args)
  (format t "~A" (nth (rand-int 0 (- (length args) 1)) args)))

(defun roll-perm (pool)
  (let* ((a-size (array-total-size pool))
	 (size (if (eq *size* nil) a-size
		 (if (<= *size* a-size) *size* a-size))))
    (loop for k from 0 to (- size 1) do
	  (let* ((i (rand-int k (- a-size 1)))
		 (r (aref pool i)))
	    (setf (aref pool i) (aref pool k))
	    (setf (aref pool k) r)
	    (format t "~A" (aref pool k))
	    (unless (= k (- size 1)) (format t "~A" *perm-separator*))))))

(defun roll-string (string-length alphabet)
  (let ((alpha-length (- (length alphabet) 1)))
    (dotimes (i string-length)
      (format t "~A" (char alphabet (rand-int 0 alpha-length))))))

(defun roll-version ()
  (format T "~A~%" *roll-version*))

(defun roll-unknown (unknown-command)
  (format *error-output*
	  "roll: command ~A not recognized, roll with no arguments for help~%"
	  unknown-command)
  (exit 1))

;;; process command line

;; extract optional integer multiplier

(if *args*
    (let ((i (parse-integer (car *args*) :junk-allowed t)))
      (if (integerp i)
	  (progn (setf *multiplier* i)
		 (shift-args 1)))))

;; parse the rest

(loop while *args* do
      (cond
					; expand pass as a str command
       ((equal "pass" (car *args*))
	   (progn
	     (assert (>= (length *args*) 2) nil
		     (format nil "roll: usage: pass <length> [<alphabet>]~%<alphabet> is combined with the default alphabet ~A" *pass-alphabet*))
	     (if (= (length *args*) 2)
		 (setf (cddr *args*) (list *pass-alphabet*))
	       (setf (caddr *args*) (concatenate 'string (caddr *args*) *pass-alphabet*)))
	     (setf (car *args*) "str")))
					; Bernoulli trial
       ((equal "b" (car *args*))
	(progn
	  (let ((p (if (>= (length *args*) 2) (parse-rational (cadr *args*)) 1/2)))
	    (assert (and (rationalp p) (<= 0 p 1))
		    nil "roll: b: optional argument must be a rational number between 0 and 1")
	    (setf *roll-command* (list 'roll-b p)))
	  (setf *args* nil)))
					; Binomial trial
       ((equal "bin" (car *args*))
	(progn
	  (assert (>= (length *args*) 3) nil "roll: usage: bin <p> <k>")
	  (let ((p (read-from-string (cadr *args*)))
		(k (parse-integer (caddr *args*))))
	    (assert (and (rationalp p) (<= 0 p 1) (integerp k))
		    nil "roll: bin: p must be a rational number between 0 and 1, k an integer.")
	    (setf *roll-command* (list 'roll-bin p k)))
	  (setf *args* nil)))
					; pick an item from a list, uniformly
       ((equal "int" (car *args*))
	(progn
	  (assert (>= (length *args*) 3) nil "roll: usage: int <from> <to>")
	  (let* ((to (parse-integer (caddr *args*)))
		 (from (parse-integer (cadr *args*)))
		 (diff (- to from)))
	    (assert (>= diff 0) nil "roll: bad range; ~A - ~A cannot be negative" to from)
	    (setf *roll-command* (list 'roll-int from to))
	    (setf *args* nil))))
					; pick an item from a list, uniformly
       ((equal "item" (car *args*))
	(progn
	  (assert (>= (length *args*) 2) nil
		  "roll: usage: item <item 1> [<item 2> <item 3> ...]")
	  (shift-args 1)
	  (expand-)
	  (assert (>= (length *args*) 1) nil
		  "roll: unexpected end of stream")
	  (setf *roll-command* (append (list 'roll-item) *args*))
	  (setf *args* nil)))
					; pick a permutation on a list, uniformly
       ((equal "perm" (car *args*))
	(progn
	  (assert (>= (length *args*) 2) nil
		  "roll: usage: perm <item 1> [<item 2> <item 3> ...]")
	  (shift-args 1)
	  (expand-)
	  (assert (>= (length *args*) 1) nil
		  "roll: unexpected end of stream")
	  (setf *roll-command*
		(list 'roll-perm (make-array (length *args*) :initial-contents *args*)))
	  (setf *args* nil)))
					; get a string on some alphabet
       ((equal "str" (car *args*))
	(progn
	  (assert (>= (length *args*) 3) nil "roll: usage: str <length> <alphabet>")
	  (assert (> (length (caddr *args*)) 0) nil "roll: alphabet cannot be empty")
	  (setf *roll-command*
		(list 'roll-string (parse-integer (cadr *args*)) (caddr *args*)))
	  (setf *args* nil)))
					; print version
       ((equal "version" (car *args*))
	(progn
	  (setf *roll-command* (list 'roll-version))
	  (setf *args* nil)))
					; set format
       ((equal "if" (car *args*))
	(progn
	  (assert (>= (length *args*) 2) nil "roll: usage: if <format>")
	  (setf *int-format* (cadr *args*))
	  (shift-args 2)))
					; set roll separator
       ((equal "rs" (car *args*))
	(progn
	  (assert (>= (length *args*) 2) nil "roll: usage: rs <separator>")
	  (setf *roll-separator* (cadr *args*))
	  (shift-args 2)))
					; set roll separator to space
       ((equal " " (car *args*))
	(progn
	  (setf *roll-separator* " ")
	  (setf *perm-separator* " ")
	  (shift-args 1)))
					; set perm separator
       ((equal "ps" (car *args*))
	(progn
	  (assert (>= (length *args*) 2) nil "roll: usage: ps <separator>")
	  (setf *perm-separator* (cadr *args*))
	  (shift-args 2)))
					; set size
       ((equal "n" (car *args*))
	(progn
	  (assert (>= (length *args*) 2) nil "roll: usage: n <integer>")
	  (setf *size* (parse-integer (cadr *args*)))
	  (shift-args 2)))
					; set source
       ((equal "src" (car *args*))
	(progn
	  (assert (>= (length *args*) 2) nil "roll: usage: src <file>")
	  (setf *random-device* (cadr *args*))
	  (shift-args 2)))
					; unknown command
       (T (progn
	    (setf *roll-command* (list 'roll-unknown (car *args*)))
	    (setf *args* nil)))))

;; no command given

(if (equal *roll-command* nil) (print-help))

;;; execute command

(with-open-file (*random-stream* *random-device* :element-type '(integer 0 255))
		(loop while (> *multiplier* 0) do (decf *multiplier*)
		      (eval *roll-command*)
		      (unless (= *multiplier* 0) (format T "~A" *roll-separator*))))
